import json
from collections.abc import Hashable
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any

import aiosqlite

from .db import AbstractConnection, AbstractDbFactory, AbstractDbManager


def _load_key(key: Any, /) -> Hashable:
    """⭐ note: unstable signature. ⭐"""
    match key:
        case Hashable():
            return key
        case list():
            return tuple(map(_load_key, key))
        case dict():
            return tuple((_load_key(k), _load_key(v)) for k, v in key.items())
        case _:
            raise TypeError("⭐ unknown json key type, cannot convert to hashable ⭐")


class Adapter(AbstractConnection):
    def __init__(self, connection: aiosqlite.Connection, db: dict[Hashable, Any]) -> None:
        self.connection = connection
        self.db = db

    def get(self, key: Hashable, default: Any) -> Any:
        return self.db.get(key, default)

    async def set(self, key: Hashable, value: Any) -> None:
        self.db[key] = value
        await self.connection.execute("REPLACE INTO kv(key, value) VALUES (?, ?)", (json.dumps(key), json.dumps(value)))
        await self.connection.commit()


@asynccontextmanager
async def manager(path: Path):
    async with aiosqlite.connect(path) as connection:
        db: dict[Hashable, Any] = {}
        await connection.execute("CREATE TABLE IF NOT EXISTS kv(key PRIMARY KEY, value)")
        for key, value in await connection.execute_fetchall("SELECT key, value FROM kv"):
            db[_load_key(json.loads(key))] = json.loads(value)
        yield Adapter(connection, db)


class Factory(AbstractDbFactory):
    def from_path(self, path: Path) -> AbstractDbManager:
        return manager(path.with_suffix(path.suffix + "sqlite"))


DBF: AbstractDbFactory = Factory()
