# rainbowadn.instrument.instrumentation

import functools
from contextlib import ExitStack
from typing import Callable, Self

__all__ = ("Instrumentation",)


class Instrumentation:
    deinstrumentation = {}
    _method: Callable
    _wrap: Callable

    def __init__(self, target, methodname: str):
        if not isinstance(methodname, str):
            raise TypeError("⭐ methodname must be str ⭐")
        if not callable(getattr(target, methodname)):
            raise TypeError("⭐ target.methodname must be callable ⭐")
        self.target = target
        self.methodname = methodname

    def instrument(self, method, *args, **kwargs):
        return method(*args, **kwargs)

    def __enter__(self) -> Self:
        if hasattr(self, "_method") or hasattr(self, "_wrap"):
            raise RuntimeError
        method = getattr(self.target, self.methodname)
        if not callable(method):
            raise TypeError
        self._method = method

        @functools.wraps(method)
        def wrap(*args, **kwargs):
            nonlocal method
            while method in self.deinstrumentation:
                self._method = method = self.deinstrumentation.pop(method)
            return self.instrument(method, *args, **kwargs)

        self._wrap = wrap

        setattr(self.target, self.methodname, wrap)

        return self

    def schedule_deinstrumentation(self):
        self.deinstrumentation[self._wrap] = self._method
        del self._wrap
        del self._method

    def deinstrument(self):
        while (method := getattr(self.target, self.methodname)) in self.deinstrumentation:
            setattr(self.target, self.methodname, self.deinstrumentation.pop(method))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.schedule_deinstrumentation()
        self.deinstrument()

    def enter(self, stack: ExitStack) -> Self:
        return stack.enter_context(self)

    def enter_conditional(self, stack: ExitStack) -> Self | None:
        if hasattr(self.target, self.methodname):
            return self.enter(stack)
        else:
            return None
