from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from time import time
from typing import AsyncContextManager, Callable, Generic, Hashable, Type, TypeVar

import discord
from discord.ext import commands

from .bot import Reservations, StarBot, StarState
from .db import AbstractConnection
from .instrumentation import Instrumentation

T = TypeVar("T")


class TypedReservations(Generic[T]):
    def __init__(self, reservations: Reservations, type_: Type[T], factory: Callable[[], T]) -> None:
        self.reservations = reservations
        self.type = type_
        self.factory = factory

    def reserve(self, key: Hashable) -> AsyncContextManager[T]:
        @asynccontextmanager
        async def inner():
            factory_called = False

            def factory():
                nonlocal factory_called
                factory_called = True
                return self.factory()

            while True:
                with self.reservations.reserve(key, factory) as reserved:
                    if not isinstance(reserved.value, self.type):
                        if factory_called:
                            raise TypeError("⭐ factory seems to have returned a value of incorrect type ⭐")
                    else:
                        yield reserved.value
                        break
                await reserved.empty()

        return inner()


class Increment:
    def __init__(self) -> None:
        self.__current = 0

    def next(self) -> int:
        self.__current += 1
        return self.__current


class EventuallyConsistent(Exception):
    pass


class AlreadyRequested(EventuallyConsistent):
    pass


class AlreadyProcessed(EventuallyConsistent):
    pass


class MessageState:
    def __init__(self) -> None:
        self.increment = Increment()
        self.requested = self.processing = self.done = self._next()
        self.__lock = asyncio.Lock()

    def _next(self):
        return self.increment.next()

    async def __aenter__(self) -> AsyncContextManager[None]:
        @asynccontextmanager
        async def inner():
            if self.processing < self.requested:
                raise AlreadyRequested()
            self.requested = current = self._next()
            async with self.__lock:
                if self.done < current:
                    self.processing = self._next()
                    yield
                    self.done = self.processing
                else:
                    raise AlreadyProcessed()

        return inner()

    async def __aexit__(self, et, _ev, _tb, /):
        return et and issubclass(et, EventuallyConsistent)


def states(state: StarState):
    return TypedReservations(state.reservations, MessageState, MessageState)


class AdminCtx:
    def __init__(self, ctx: commands.Context) -> None:
        self.ctx = ctx
        assert ctx.guild
        self.guild: discord.Guild = ctx.guild
        self.bot: StarBot = ctx.bot
        self.state: StarState = self.bot.starstate
        self.connection: AbstractConnection = self.state.connection

    async def assign(self, count: int) -> None:
        await self.connection.set(("assign", self.guild.id), {"channel": self.ctx.channel.id, "count": count})

    async def unassign(self) -> None:
        await self.connection.set(("assign", self.guild.id), None)


class StarForwardCtx:
    def __init__(self, message: discord.Message, star_channel: discord.abc.Messageable) -> None:
        self.message = message
        self.star_channel = star_channel

    async def on(self) -> None:
        author = self.message.author
        embed = discord.Embed(description=self.message.content or None)
        avatar = author.avatar
        embed.set_author(name=author.display_name, url=self.message.jump_url, icon_url=avatar and avatar.url)
        image = next(
            (
                attachment.url
                for attachment in self.message.attachments
                if (attachment.content_type or "").startswith("image/")
            ),
            None,
        )
        if image is not None:
            embed.set_image(url=image)
        await asyncio.gather(
            self.message.add_reaction("⭐"),
            self.star_channel.send(embed=embed),
        )


class StarMessageCtx:
    def __init__(self, message: discord.Message, star_channel: discord.abc.Messageable, count: int) -> None:
        self.message = message
        self.star_channel = star_channel
        self.count = count

    def triggered(self) -> bool:
        reaction = next((reaction for reaction in self.message.reactions if reaction.emoji == "⭐"), None)
        return reaction is not None and not reaction.me and reaction.count >= self.count

    async def on(self) -> None:
        if self.triggered():
            await StarForwardCtx(self.message, self.star_channel).on()


class StarEventCtx:
    def __init__(self, reaction: ReactionCtx, star_channel_id: int, count: int) -> None:
        self.bot = reaction.bot
        self.states = states(reaction.state)
        self.star_channel_id = star_channel_id
        self.count = count
        self.channel_id = reaction.channel_id
        self.message_id = reaction.message_id

    async def _get_channel(self, id_: int) -> discord.abc.Messageable:
        channel = self.bot.get_channel(id_) or await self.bot.fetch_channel(id_)
        match channel:
            case discord.CategoryChannel() | discord.ForumChannel() | discord.abc.PrivateChannel():
                raise TypeError
            case _:
                return channel

    async def _get_message(self) -> discord.Message:
        event_channel = await self._get_channel(self.channel_id)
        return await event_channel.fetch_message(self.message_id)

    async def _on(self) -> None:
        star_channel, message = await asyncio.gather(self._get_channel(self.star_channel_id), self._get_message())
        await StarMessageCtx(message, star_channel, self.count).on()

    async def on(self) -> None:
        async with self.states.reserve(self.message_id) as state, state as guard, guard:
            await asyncio.create_task(self._on())


class ReactionCtx:
    def __init__(self, bot: StarBot, event: discord.RawReactionActionEvent) -> None:
        self.bot = bot
        self.guild_id = event.guild_id
        self.channel_id = event.channel_id
        self.message_id = event.message_id
        self.name = event.emoji.name
        self.state: StarState = self.bot.starstate
        self.connection: AbstractConnection = self.state.connection

    async def on(self) -> None:
        if self.name != "⭐":
            return
        match self.connection.get(("assign", self.guild_id), None):
            case {"channel": int() as star_channel_id, "count": int() as count}:
                pass
            case _:
                return
        if self.channel_id == star_channel_id:
            return
        await StarEventCtx(self, star_channel_id, count).on()


class Stars(commands.Cog):
    def __init__(self, bot: StarBot) -> None:
        super().__init__()
        self.__bot = bot

    @commands.hybrid_command()
    async def ping(self, ctx: commands.Context):
        print("⭐ ping pong ⭐")
        await ctx.reply("pong", mention_author=False)

    @commands.hybrid_command()
    @commands.is_owner()
    async def reload(self, ctx: commands.Context):
        print("⭐ reload ⭐")
        bot: commands.Bot = ctx.bot
        try:
            await bot.reload_extension("starbot.stars")
        except commands.ExtensionNotLoaded:
            await ctx.reply("⭐ not loaded ⭐")
        print("⭐ reloaded ⭐")
        await ctx.reply("⭐ reloaded ⭐")

    @commands.hybrid_command()
    @commands.is_owner()
    async def sync(self, ctx: commands.Context):
        await ctx.bot.tree.sync()
        print("⭐ synced ⭐")
        await ctx.reply("⭐ synced ⭐")

    @commands.hybrid_command()
    @commands.has_permissions(administrator=True)
    async def assign(self, ctx: commands.Context, count: int):
        await AdminCtx(ctx).assign(count)
        await ctx.reply("⭐ assigned ⭐")

    @commands.hybrid_command()
    @commands.has_permissions(administrator=True)
    async def unassign(self, ctx: commands.Context):
        await AdminCtx(ctx).unassign()
        await ctx.reply("⭐ unassigned ⭐")

    @commands.Cog.listener()
    async def on_raw_reaction_add(self, event: discord.RawReactionActionEvent):
        await ReactionCtx(self.__bot, event).on()

    @commands.hybrid_command()
    @commands.is_owner()
    async def timeit(self, ctx: commands.Context, duration: float):
        with Timeit(ReactionCtx, "on") as timeit:
            await asyncio.sleep(duration)
        await ctx.reply(f"```\n{timeit.times}\n```")

    @commands.hybrid_command()
    @commands.is_owner()
    async def throttle(self, ctx: commands.Context, duration: float, delay: float):
        with Cooldown(StarEventCtx, "_on", delay):
            await asyncio.sleep(duration)
        await ctx.reply(f"⭐ done ⭐")


class Cooldown(Instrumentation):
    def __init__(self, target, methodname: str, delay: float):
        super().__init__(target, methodname)
        self.delay = delay

    async def instrument(self, method, *args, **kwargs):
        await method(*args, **kwargs)
        await asyncio.sleep(self.delay)


class Timeit(Instrumentation):
    def __init__(self, target, methodname: str):
        super().__init__(target, methodname)
        self.times = []

    async def instrument(self, method, *args, **kwargs):
        t = time()
        await method(*args, **kwargs)
        self.times.append((t, time()))


async def setup(bot: StarBot):
    global cog
    cog = Stars(bot)
    await bot.add_cog(cog)
    print("⭐ set up ⭐")


async def teardown(bot: StarBot):
    global cog
    await bot.remove_cog(cog.qualified_name)
    del cog
    print("⭐ torn down ⭐")
